@@ -8,6 +8,7 @@ import os
 import time
 import errno
 import string
+import urllib.request
 from yarl import URL
 from http.cookies import SimpleCookie, BaseCookie
 
@@ -18,12 +19,19 @@ import bs4
 from ._common import log, kw_only
 from . import _graphql, _util, _exception
 
-from typing import Optional, Mapping, Callable, Any, Awaitable
+from typing import Optional, Mapping, Callable, Any, Awaitable, Dict, List, NamedTuple
+
+try:
+    from aiohttp_socks import ProxyType, ProxyConnector
+except ImportError:
+    ProxyType = None
+    ProxyConnector = None
 
 SERVER_JS_DEFINE_REGEX = re.compile(r'(?:'
                                     r'\(new ServerJS\(\)\)(?:;s)?'
                                     r'|\(require\("ServerJS(?:Define)?"\)\)\(\)'
-                                    r').handle(?:Defines)?\(')
+                                    r').handle(?:Defines|WithCustomApplyEach)?\('
+                                    r'(?:ScheduledApplyEach,)?')
 SERVER_JS_DEFINE_JSON_DECODER = json.JSONDecoder()
 
 
@@ -53,9 +61,9 @@ def parse_server_js_define(html: str) -> Mapping[str, Any]:
     if not define_splits:
         file_name = write_html_to_temp(html)
         raise _exception.ParseError("Could not find any ServerJSDefine", data_file=file_name)
-    if len(define_splits) > 2:
-        file_name = write_html_to_temp(html)
-        raise _exception.ParseError("Found too many ServerJSDefine", data_file=file_name)
+    # if len(define_splits) > 2:
+    #     file_name = write_html_to_temp(html)
+    #     raise _exception.ParseError("Found too many ServerJSDefine", data_file=file_name)
     try:
         parsed, _ = SERVER_JS_DEFINE_JSON_DECODER.raw_decode(define_splits[0], idx=0)
     except json.JSONDecodeError as e:
@@ -77,6 +85,52 @@ def parse_server_js_define(html: str) -> Mapping[str, Any]:
     return _util.get_jsmods_define(rtn)
 
 
+def parse_kv(vals: List[str]) -> Dict[str, str]:
+    kv = {}
+    for val in vals:
+        split = val.strip().split("=", 1)
+        if len(split) == 1:
+            kv[split[0]] = True
+        else:
+            kv[split[0]] = split[1]
+    return kv
+
+
+class AltSvc(NamedTuple):
+    alt_authority: str
+    max_age: int
+    persist: bool
+    extra_meta: Dict[str, str]
+
+
+def parse_alt_svc(r: aiohttp.ClientResponse) -> Dict[str, AltSvc]:
+    try:
+        header = r.headers["Alt-Svc"]
+    except KeyError:
+        return {}
+    if header.lower() == "clear":
+        return {}
+    services = {}
+    for service in header.split(","):
+        vals = service.split(";")
+        try:
+            protocol_id, alt_authority = vals[0].split("=")
+        except ValueError:
+            continue
+        alt_authority: str = alt_authority.strip('"')
+        kv = parse_kv(vals[1:])
+        try:
+            max_age = int(kv.pop("max_age"))
+        except (KeyError, ValueError):
+            max_age = 86400
+        try:
+            persist = kv.pop("persist") == "1"
+        except KeyError:
+            persist = False
+        services[protocol_id] = AltSvc(alt_authority, max_age, persist, extra_meta=kv)
+    return services
+
+
 def base36encode(number: int) -> str:
     """Convert from Base10 to Base36."""
     # Taken from https://en.wikipedia.org/wiki/Base36#Python_implementation
@@ -93,21 +147,15 @@ def base36encode(number: int) -> str:
     return sign + result
 
 
-def prefix_url(url: str) -> str:
-    if url.startswith("/"):
-        return "https://www.messenger.com" + url
-    return url
-
-
 def generate_message_id(now: datetime.datetime, client_id: str) -> str:
     k = _util.datetime_to_millis(now)
     l = int(random.random() * 4294967295)
     return "<{}:{}-{}@mail.projektitan.com>".format(k, l, client_id)
 
 
-def get_user_id(session: aiohttp.ClientSession) -> str:
+def get_user_id(domain: str, session: aiohttp.ClientSession) -> str:
     try:
-        rtn = session.cookie_jar.filter_cookies(URL("https://messenger.com")).get("c_user")
+        rtn = session.cookie_jar.filter_cookies(URL(f"https://{domain}")).get("c_user")
     except (AttributeError, KeyError):
         raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
     if rtn is None:
@@ -115,21 +163,19 @@ def get_user_id(session: aiohttp.ClientSession) -> str:
     return rtn if isinstance(rtn, str) else str(rtn.value)
 
 
-class ProxyClientSession(aiohttp.ClientSession):
-    def __init__(self, *args, **kwargs) -> None:
-        super().__init__(*args, **kwargs)
-        self._proxy = os.environ.get("HTTP_PROXY")
-
-    def _request(self, *args, **kwargs) -> Awaitable[aiohttp.ClientResponse]:
-        return super()._request(*args, **kwargs, proxy=self._proxy)
-
-
-def session_factory(user_agent: Optional[str] = None) -> aiohttp.ClientSession:
+def session_factory(domain: str, user_agent: Optional[str] = None) -> aiohttp.ClientSession:
     from . import __version__
-    return ProxyClientSession(headers={
-        "Referer": "https://www.messenger.com/",
-        "User-Agent": user_agent or f"fbchat-asyncio/{__version__}",
-    })
+    try:
+        http_proxy = urllib.request.getproxies()["http"]
+    except KeyError:
+        http_proxy = None
+    return aiohttp.ClientSession(connector=(ProxyConnector.from_url(http_proxy)
+                                            if ProxyConnector and http_proxy
+                                            else None),
+                                 headers={
+                                     "Referer": f"https://www.{domain}/",
+                                     "User-Agent": user_agent or f"fbchat-asyncio/{__version__}",
+                                 })
 
 
 def login_cookies(at: datetime.datetime):
@@ -237,9 +283,17 @@ def get_fb_dtsg(define) -> Optional[str]:
         return define["DTSGInitData"]["token"]
     elif "DTSGInitialData" in define:
         return define["DTSGInitialData"]["token"]
+    elif "MRequestConfig" in define and "dtsg" in define["MRequestConfig"]:
+        return define["MRequestConfig"]["dtsg"]["token"]
     return None
 
 
+def prefix_url(domain: str, path: str) -> URL:
+    if path.startswith("/"):
+        return URL(f"https://www.{domain}" + path)
+    return URL(path)
+
+
 @attr.s(slots=True, kw_only=kw_only, repr=False, eq=False, auto_attribs=True)
 class Session:
     """Stores and manages state required for most Facebook requests.
@@ -250,10 +304,15 @@ class Session:
     _user_id: str
     _fb_dtsg: str
     _revision: int
+    domain: str
+    _onion: Optional[str] = None
     _session: aiohttp.ClientSession = attr.ib(factory=session_factory)
     _counter: int = 0
     _client_id: str = attr.ib(factory=client_id_factory)
 
+    def _prefix_url(self, path: str) -> URL:
+        return prefix_url(self.domain, path)
+
     @property
     def user(self):
         """The logged in user."""
@@ -268,7 +327,7 @@ class Session:
         return "<fbchat.Session user_id={}>".format(self._user_id)
 
     def _get_params(self):
-        self._counter += 1  # TODO: Make this operation atomic / thread-safe
+        self._counter += 1
         return {
             "__a": 1,
             "__req": base36encode(self._counter),
@@ -276,7 +335,6 @@ class Session:
             "fb_dtsg": self._fb_dtsg,
         }
 
-    # TODO: Add ability to load previous cookies in here, to avoid 2fa flow
     @classmethod
     async def login(cls, email: str, password: str,
                     on_2fa_callback: Callable[[], Awaitable[int]] = None,
@@ -311,7 +369,7 @@ class Session:
             >>> session.user.id
             "1234"
         """
-        session = session_factory(user_agent=user_agent)
+        session = session_factory(domain="messenger.com", user_agent=user_agent)
 
         data = {
             # "jazoest": "2754",
@@ -379,7 +437,7 @@ class Session:
             raise _exception.NotLoggedIn("Failed logging in: {}, {}".format(url, error))
 
         try:
-            return await cls._from_session(session=session)
+            return await cls._from_session(session=session, domain="messenger.com")
         except _exception.NotLoggedIn as e:
             raise _exception.ParseError("Failed loading session", data=r) from e
 
@@ -394,12 +452,16 @@ class Session:
         """
         # Send a request to the login url, to see if we're directed to the home page
         try:
-            r = await self._session.get(prefix_url("/login/"), allow_redirects=False)
+            r = await self._session.get(self._prefix_url("/login/"), allow_redirects=False)
         except aiohttp.ClientError as e:
             _exception.handle_requests_error(e)
             raise Exception("handle_requests_error did not raise exception")
         _exception.handle_http_error(r.status)
-        return "https://www.messenger.com/" == r.headers.get("Location")
+        location = r.headers.get("Location")
+        return location in (f"https://www.{self.domain}/",
+                            # We include this as a "logged in" status, since the user is logged in,
+                            # but needs to verify the session elsewhere
+                            f"https://www.{self.domain}/checkpoint/block/")
 
     async def logout(self) -> None:
         """Safely log out the user.
@@ -409,10 +471,13 @@ class Session:
         Example:
             >>> session.logout()
         """
+        if self.domain != "messenger.com":
+            log.warning("session.logout() is only supported on messenger.com")
+            return
         data = {"fb_dtsg": self._fb_dtsg}
         try:
             r = await self._session.post(
-                prefix_url("/logout/"), data=data, allow_redirects=False
+                self._prefix_url("/logout/"), data=data, allow_redirects=False
             )
         except aiohttp.ClientError as e:
             _exception.handle_requests_error(e)
@@ -427,19 +492,26 @@ class Session:
             )
 
     @classmethod
-    async def _from_session(cls, session: aiohttp.ClientSession) -> Optional['Session']:
+    async def _from_session(cls, session: aiohttp.ClientSession, domain: str
+                            ) -> Optional['Session']:
         # TODO: Automatically set user_id when the cookie changes in the session
-        user_id = get_user_id(session)
+        user_id = get_user_id(domain, session)
 
         # Make a request to the main page to retrieve ServerJSDefine entries
         try:
-            r = await session.get(prefix_url("/"), allow_redirects=False)
+            r = await session.get(prefix_url(domain, "/"), allow_redirects=False, headers={
+                "Accept": "text/html",
+            })
         except aiohttp.ClientError as e:
             _exception.handle_requests_error(e)
             raise Exception("handle_requests_error did not raise exception")
         _exception.handle_http_error(r.status)
 
-        define = parse_server_js_define(await r.text())
+        html = await r.text()
+        if len(html) == 0:
+            raise _exception.FacebookError("Got empty response when trying to check login")
+
+        define = parse_server_js_define(html)
 
         fb_dtsg = get_fb_dtsg(define)
         if fb_dtsg is None:
@@ -449,13 +521,19 @@ class Session:
             raise _exception.NotLoggedIn(
                 "Found empty fb_dtsg, the session was probably invalid."
             )
-
         try:
             revision = int(define["SiteData"]["client_revision"])
         except TypeError:
             raise _exception.ParseError("Could not find client revision", data=define)
+        onion = None
+        alt_svc_data = parse_alt_svc(r)
+        if "h2" in alt_svc_data and alt_svc_data["h2"].alt_authority.endswith(".onion:443"):
+            # TODO remember expiry too?
+            onion = alt_svc_data["h2"].alt_authority
+            log.info("Got onion alt-svc %s", onion)
 
-        return cls(user_id=user_id, fb_dtsg=fb_dtsg, revision=revision, session=session)
+        return cls(user_id=user_id, fb_dtsg=fb_dtsg, revision=revision, session=session,
+                   domain=domain, onion=onion)
 
     def get_cookies(self) -> Optional[Mapping[str, str]]:
         """Retrieve session cookies, that can later be used in `from_cookies`.
@@ -466,12 +544,12 @@ class Session:
         Example:
             >>> cookies = session.get_cookies()
         """
-        cookie = self._session.cookie_jar.filter_cookies(URL("https://messenger.com"))
+        cookie = self._session.cookie_jar.filter_cookies(URL(f"https://{self.domain}"))
         return {key: morsel.value for key, morsel in cookie.items()}
 
     @classmethod
-    async def from_cookies(cls, cookies: Mapping[str, str], user_agent: Optional[str] = None
-                           ) -> 'Session':
+    async def from_cookies(cls, cookies: Mapping[str, str], user_agent: Optional[str] = None,
+                           domain: str = "messenger.com") -> 'Session':
         """Load a session from session cookies.
 
         Args:
@@ -482,7 +560,7 @@ class Session:
             >>> # Store cookies somewhere, and then subsequently
             >>> session = fbchat.Session.from_cookies(cookies)
         """
-        session = session_factory(user_agent=user_agent)
+        session = session_factory(domain=domain, user_agent=user_agent)
 
         if isinstance(cookies, BaseCookie):
             cookie = cookies
@@ -490,10 +568,10 @@ class Session:
             cookie = SimpleCookie()
             for key, value in cookies.items():
                 cookie[key] = value
-                cookie[key].update({"domain": "messenger.com", "path": "/"})
-        session.cookie_jar.update_cookies(cookie, URL("https://messenger.com"))
+                cookie[key].update({"domain": domain, "path": "/"})
+        session.cookie_jar.update_cookies(cookie, URL(f"https://{domain}"))
 
-        return await cls._from_session(session=session)
+        return await cls._from_session(session=session, domain=domain)
 
     async def _post(self, url, data, files=None, as_graphql=False):
         data.update(self._get_params())
@@ -504,8 +582,17 @@ class Session:
             for key, (name, file, content_type) in files.items():
                 payload.add_field(key, file, filename=name, content_type=content_type)
             data = payload
+        real_url = self._prefix_url(url)
+        kwargs = {}
+        if self._onion:
+            # TODO is there some way to change the host aiohttp connects to without changing the
+            #      domain it uses for TLS, cookies and the Host header?
+            kwargs["ssl"] = False
+            kwargs["headers"] = {"Host": real_url.host}
+            kwargs["cookies"] = self._session.cookie_jar.filter_cookies(real_url)
+            real_url = real_url.with_host(real_url.host.replace(self.domain, self._onion))
         try:
-            r = await self._session.post(prefix_url(url), data=data)
+            r = await self._session.post(real_url, data=data, **kwargs)
         except aiohttp.ClientError as e:
             _exception.handle_requests_error(e)
             raise Exception("handle_requests_error did not raise exception")
